import os 
from PIL import Image 
import numpy as np 
import cv2

# Creating a generator so I can save storage in the RAM 
def load_images(path): #Either single file or folder 
    if os.path.isfile(path):
        image = Image.open(path)
        yield path, image 

    else:
        for file_name in os.listdir(path): #Assuming this is the string of the folder containing images 
            if file_name.lower().endswith((".jpg", ".png", ".jpeg")):
                full_path = os.path.join(path, file_name)
                image = Image.open(full_path)
                yield full_path, image 

def mask_to_polygon(mask: np.ndarray, epsilon_ratio=0.005): #Convert binary mask to polygon using OpenCV contours
    mask = mask.astype(np.uint8)

    contours, _ = cv2.findContours(
        mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE
    )

    if not contours:
        return None

    # Take the largest contour
    contour = max(contours, key=cv2.contourArea)

    # Polygon simplification
    epsilon = epsilon_ratio * cv2.arcLength(contour, True)
    polygon = cv2.approxPolyDP(contour, epsilon, True)

    return polygon.squeeze(1)  # (N, 2)

def mask_to_contour_format(masks, image):
    array_img = np.asarray(image)
    image_h, image_w, _ = array_img.shape
    lines = []
    masks = masks.cpu().numpy()

    for i in range(masks.shape[0]):
        # Pick best mask (largest area)
        mask_set = masks[i]
        areas = [m.sum() for m in mask_set]
        best_mask = mask_set[np.argmax(areas)]

        polygon = mask_to_polygon(best_mask)
        if polygon is None or len(polygon) < 3:
            continue

        # Normalize coordinates
        polygon = polygon.astype(float)
        polygon[:, 0] /= image_w
        polygon[:, 1] /= image_h

        polygon = polygon.clip(0, 1)

        # Flatten
        poly_flat = polygon.reshape(-1)
        poly_str = " ".join(f"{p:.6f}" for p in poly_flat)

        #line = f"{class_ids[i]} {poly_str}"
        #lines.append(line)
        return poly_str


import torch
from PIL import Image
from transformers import AutoProcessor, AutoModelForZeroShotObjectDetection
from accelerate import Accelerator
import logging
import time
import os
from pathlib import Path

logging.basicConfig(level=logging.INFO)

# __init__ to load the model only once
# Will only load the grounding dino model when the class is called and not during the import of this file 
class GroundingDINOTiny:
    def __init__(self, model_id="IDEA-Research/grounding-dino-tiny", cache_dir="./cache"): 
        self.model_id = model_id
        self.cache_dir = cache_dir

        self.device = Accelerator().device

        self.processor = AutoProcessor.from_pretrained(self.model_id, use_fast=False,
                                                       cache_dir=self.cache_dir)
        self.model = AutoModelForZeroShotObjectDetection.from_pretrained(
            self.model_id,
            cache_dir=self.cache_dir).to(self.device)

    def detect(self, image, text_labels, image_path, save_results=False, results_folder = "detector_results", box_threhsold=0.4, text_threshold=0.3):
        inputs = self.processor(images=image, text=text_labels, return_tensors="pt").to(self.device)
        
        with torch.no_grad():
            outputs = self.model(**inputs)

        results = self.processor.post_process_grounded_object_detection(
            outputs,
            inputs.input_ids,
            threshold=box_threhsold,
            text_threshold=text_threshold,
            target_sizes=[image.size[::-1]]
            )
        if save_results: #Only if flag is True for saving results
            self.save_results_txt(results, results_folder, image_path)
        return results
    
    def save_results_txt(self, results, results_folder, image_path):
        if results: #There are detections and no empty detections there
            os.makedirs(results_folder, exist_ok=True)
            file_name = Path(image_path).name.split(".")[0]
            label_path = os.path.join(results_folder, file_name+".txt")
            bboxes = results[0]["boxes"].cpu().numpy()
            conf_scores = results[0]["scores"].cpu().numpy()
            text_labels = results[0]["text_labels"]
            labels = results[0]["labels"] #Should return integer labels in the future as per the warnings currently 
            for index, box in enumerate(bboxes):
                x1, y1, x2, y2 = map(int, box)
                confidence_score = conf_scores[index]
                text_class = text_labels[index]
                with open(label_path, "a") as f:
                    line = text_class + "\t" + str(x1) + "\t" + str(y1) + "\t" + str(x2) + "\t" + str(y2) + str(confidence_score)
                    f.write(line + "\n")
        else:
            logging.info(f"Detections missed by Grounding Dino Tiny\t{time.time()}")


from transformers import Sam2Processor, Sam2Model
import torch
from PIL import Image
import numpy as np 
# import matplotlib.pyplot as plt
import cv2
from vision_loom.models.grounding_dino import GroundingDINOTiny
import logging
from vision_loom.utils.io import mask_to_polygon, mask_to_contour_format
import os
from pathlib import Path 
from tqdm import tqdm

logging.basicConfig(level=logging.INFO)

class SAM2:
    def __init__(self, model_id="facebook/sam2.1-hiera-tiny", device_map="auto", cache_dir = "./cache"):
        self.model_id = model_id
        self.device_map = device_map 
        self.cache_dir = cache_dir

        model = Sam2Model.from_pretrained(self.model_id,
                                  device_map = self.device_map,
                                  cache_dir=self.cache_dir)
        self.model = model.to(model.device)
        logging.info("SAM2 Model Loaded")
        
        self.processor = Sam2Processor.from_pretrained(self.model_id,
                                          use_fast=True,
                                          cache_dir = self.cache_dir)
        logging.info("SAM2 Processor Loaded")

    def detect(self, image, text_labels, image_path, save_results=False, results_folder="detector_results"):
        detector_model = GroundingDINOTiny()
        detector_results = detector_model.detect(image, text_labels, image_path)
        if detector_results:
            bboxes = detector_results[0]["boxes"].cpu().numpy()
            conf_scores = detector_results[0]["scores"].cpu().numpy()
            text_labels = detector_results[0]["text_labels"]
            labels = detector_results[0]["labels"] #Should return integer labels in the future as per the warnings currently 
            for index, box in tqdm(enumerate(bboxes)):
                x1, y1, x2, y2 = map(int, box)
                confidence_score = conf_scores[index]
                text_class = text_labels[index]
                input_boxes = [[[x1, y1, x2, y2]]]
                inputs = self.processor(images=image, input_boxes=input_boxes, return_tensors="pt").to(self.model.device)
                with torch.no_grad():
                    outputs = self.model(**inputs)
                masks = self.processor.post_process_masks(outputs.pred_masks.cpu(), inputs["original_sizes"])[0] #masks of one object of one class currently
                poly_lines = mask_to_contour_format(masks, image)
                if save_results:
                    self.save_results(poly_lines, text_class, confidence_score, results_folder, image_path)
        else:
            logging.info("No object found from prompt labels")
            return None 
        
    def save_results(self, poly_lines, text_class, confidence_score, results_folder, image_path):
            os.makedirs(results_folder, exist_ok=True)
            file_name = Path(image_path).name.split(".")[0]
            label_path = os.path.join(results_folder, file_name+".txt")
            results = str(text_class) + "\t" + poly_lines + "\t" + str(confidence_score)

            with open(label_path, "a") as f:
                f.write(results + "\n")